""" General utilities/helper functions"""
import re
from collections.abc import MutableMapping


# individual cell utilities
def strip_html(html_string):
    if html_string:
        return re.sub(r"<[^>]+>", "", html_string)
    else:
        return html_string


def to_int_if_base10(val):
    """
    converts value to a string and if
    float (or a string rep of a float) of base10
    to an integer string representation.

    NOTE:
    """
    string = str(val)

    if "." in string:
        parts = string.split(".")
        if len(parts) == 2 and parts[1] == "0":
            return parts[0]

    return string


def parse_dictionary_str(string, item_sep, keyval_sep):
    """
    parses a stringified dictionary into a dictionary
    based on item separator

    """
    stritems = string.strip().split(item_sep)
    items = {}
    for stritem in stritems:
        item = stritem.split(keyval_sep, 1)
        items[item[0]] = item[1].strip()

    return items


def parse_list_str(string, list_sep):
    return string.strip().split(list_sep)


# dictionary utilities
def flatten_except_if(dictionary, parent_key=False, sep=".", except_keys=["encodings"]):
    """
    Turn a nested dictionary into a flattened dictionary. Taken from gen3
    mds.agg_mds.adapter.flatten
    but added except keys and fixed a bug where parent is always False in MutableMapping

    :param dictionary: The dictionary to flatten
    :param parent_key: The string to prepend to dictionary's keys
    :param sep: The string used to separate flattened keys
    :param except_keys: keys to not flatten. Note, can be nested if using notation specified in sep
    :return: A flattened dictionary
    """

    items = []
    for key, value in dictionary.items():
        new_key = str(parent_key) + sep + key if parent_key else key
        if isinstance(value, MutableMapping) and not new_key in except_keys:
            items.extend(flatten_except_if(value, new_key, sep).items())
        else:
            items.append((new_key, value))
    return dict(items)

def stringify_keys(dictionary):
    orig_keys = list(dictionary.keys())
    for key in orig_keys:
        dictionary[str(key)] = dictionary.pop(key)

def convert_rec_to_json(field):
    """
    converts a flattened dictionary to a nested dictionary
    based on JSON path dot notation indicating nesting
    """
    field_json = {}
    for prop_path, prop in field.items():
        if str(prop) and str(prop) != "<NA>" and str(prop) != "nan":
            # initiate the prop to be added with the entire
            # field
            prop_json = field_json
            # get the inner most dictionary item of the jsonpath
            nested_names = prop_path.split(".")
            for i, prop_name in enumerate(nested_names):
                is_last_nested = i + 1 == len(nested_names)
                if prop_json.get(prop_name) and not is_last_nested:
                    prop_json = prop_json[prop_name]
                # if no object currently
                elif not is_last_nested:
                    prop_json[prop_name] = {}
                    prop_json = prop_json[prop_name]
                # assign property to inner most item
                else:
                    prop_json[prop_name] = prop

    return field_json


# documentation building utilities
def find_docstring_desc(fxn):
    """
    return the description part of a docstring
    (ie text before Parameters)
    """
    exp = "^(.*)Parameters\\n"

    if fxn.__doc__:
        docstring = fxn.__doc__.strip()
    else:
        docstring = "No documentation"
    try:
        return re.search(exp, docstring, re.DOTALL).group(1)
    except AttributeError:
        return docstring


# add missing values and order according to the order of a list


def sync_fields(data, field_list,missing_value=None):
    """
    Sorts fields and adds missing fields (with None value).
    If extra fields exist in a record that are not in field_list, then tacks on at 
    end of record.


    Parameters
    --------------
    data [list]: json array of values
    fields [list]: the list of all fields (e.g., from a schema)

    Returns
    -------------
    list: json array with fields added if missing
    """
    data_with_missing = []

    for record in data:
        extra_fields = list(set(list(record)).difference(field_list))
        new_record = {field:record.get(field, missing_value) 
            for field in field_list+extra_fields}
        data_with_missing.append(new_record)

    
    return data_with_missing



# join mappings for json to csv

def join_iter(iterable,sep_list="|"):
    return sep_list.join([str(p) for p in iterable])

def join_dictvals(dictionary:dict,sep:str):
    return sep.join(dictionary.values())

def join_dictitems(dictionary:dict,sep_keyval='=',sep_items='|'):
    """ joins a mappable collection (ie dictionary) into a string
    representation with specified separators for the key and value
    in addition to items. 

    All items are coerced to the string representation (eg if key or value
    is None, this will be coerced to "None")
    """
    dict_list = []
    for key,val in dictionary.items():
        keystr = str(key)
        valstr = str(val)
        dict_list.append(keystr+sep_keyval+valstr)
    return sep_items.join(dict_list)


joinmap = {
    'constraints.enum': join_iter,
    'cde_id': join_dictvals,
    'ontology_id': join_dictvals,
    'encodings': join_dictitems,
    'missingValues':join_iter,
    'trueValues': join_iter,
    'falseValues':join_iter,
    # TODO: add stats
}

def join_prop(propname,prop):
    return joinmap[propname](prop) if propname in joinmap else prop


from healdata_utils.utils import sync_fields


def test_add_missing_fields():
    # data (note one var not in schema)
    data = [
        {"var2": 2, "var3": 3, "var1": 1},
        {
            "var5": 10,
            "var4": 9,
            "var1": 1,
        },
    ]
    # fields from schema
    fields = ["var1", "var2", "var3", "var4"]

    data_with_missing = sync_fields(data, fields)
    assert data_with_missing == [
        {"var1": 1, "var2": 2, "var3": 3, "var4": None},
        {"var1": 1, "var2": None, "var3": None, "var4": 9, "var5": 10}
    ]


''' 
takes the filled out variable level metadata template
and converts to json, validates, and saves json template to file
based on an explicit set of mapping functions

goes the other way (from json to csv as well)

see convert_templatecsv_to_json and convert_json_to_templatecsv
''' 
import petl as etl
from pathlib import Path
# from frictionless import Resource,Package
from healdata_utils.utils import convert_rec_to_json
from healdata_utils.io import read_table
from .mappings import fieldmap,zipmap,typemap,castnumbers
from os import PathLike

def convert_templatecsv(
    csvtemplate: str,
    data_dictionary_props: dict,
    mappings: dict = fieldmap,
) -> dict:
    """
    Converts a CSV conforming to HEAL specifications (but see 2 additional notes below) 
    into a HEAL-specified data dictionary in both csv format and json format.

    Converts an in-memory data dictionary or a path to a data dictionary file into a HEAL-specified tabular template by:
        1. Adding missing fields, and
        2. Converting fields from a specified mapping.
            NOTE: currently this mapping is only float/num to number or text/char to string (case insensitive)
                In future versions, there will be a specified module for csv input mappings.
    
    Parameters
    ----------
    csvtemplate : str or path-like or an object that can be inferred as data by frictionless's Resource class.
        Data or path to data with the data being a tabular HEAL-specified data dictionary.
        This input can be any data object or path-like string excepted by a frictionless Resource object.
    data_dictionary_props : dict
        The HEAL-specified data dictionary properties.
    mappings : dict, optional
        Mappings (which can be a dictionary of either lambda functions or other to-be-mapped objects).
        Default: specified fieldmap.

    Returns
    -------
    dict
        A dictionary with two keys:
            - 'templatejson': the HEAL-specified JSON object.
            - 'templatecsv': the HEAL-specified tabular template.

    """

    if isinstance(csvtemplate,(str,PathLike)):
        template_tbl = read_table(str(Path(csvtemplate)))
    else:
        template_tbl = pd.DataFrame(csvtemplate)

    # apply convert functions for fields that exist in input
    convertfields = {
        propname:fxn 
        for propname,fxn in mappings.items() 
        if propname in template_tbl
    }
    castfields = {
        propname:fxn 
        for propname,fxn in castnumbers.items() 
        if propname in template_tbl
    }

    tbl_csv = (
        etl.fromdataframe(template_tbl)
        .convert(castfields)
        .convertall({None:""})
    )
    fields_csv = tbl_csv.dicts()
    tbl_json = (
        tbl_csv
        .convert(convertfields)
    )

    fields_json = []
    for record in tbl_json.dicts():
        jsonrecord = convert_rec_to_json(record)
        # if supposed to be an array of records,with properties spanning multiple columns in templatecsv, create that
        # eg {"test":[1,2]} --> [{"test":1},{"test":2}]
        for name in zipmap:
            if name in jsonrecord:
                field = jsonrecord.pop(name)
                headers = field.keys()
                values = field.values()
                newfield = etl.fromcolumns(values,header=headers).dicts()
                jsonrecord[name] = [{name:val for name,val in _record.items() if val} 
                    for _record in newfield]
        
        fields_json.append(jsonrecord)

    template_json = dict(**data_dictionary_props,data_dictionary=fields_json)
    template_csv = dict(**data_dictionary_props,data_dictionary=fields_csv)

    return {"templatejson":template_json,"templatecsv":template_csv}


''' 
contains mappings (both lambda functions or column mappings)
''' 

from healdata_utils import schemas
# split array columns
def split_str_array(string,sep='|'):
    if string:
        return [s.strip() for s in string.split(sep)]
    else:
        return string

# if object within array, assign to properties
def map_keys_vals(keys,vals):
    ''' zips two lists of the same size as 
    a dictionary
    ''' 
    return dict(zip(keys,vals))

def split_and_map(string,prop):
    ''' 
    splits nested stringified delimited lists 
    (delimiters being | for outer and = for inner)
    and zips/maps each of the inner lists to a set
    of values (right now keys of a dictionary)
    TODO: rename function split_and_map_to_keys
    TODO: generalize to more than keys


    '''
    if string:
        keys = prop['items']['properties'].keys()
        return [
            map_keys_vals(keys,split_str_array(x,sep='=')) 
            for x in split_str_array(string,sep='|')
        ]
    else:
        return string

def loads_dict(string,item_sep='|',key_val_sep='='):
    if string:
        return dict([split_str_array(s,key_val_sep) 
            for s in split_str_array(string,item_sep)])
    else:
        return string
def mapval(v,mapping):
    v = str(v)
    if v in mapping:
        return mapping[v]
    else:
        return v

def to_bool(v):
    if v.lower() in true_values:
        return True 
    elif v.lower() in false_values:
        return False 
    else:
        return ""
        
typemap = {
    #from bacpac
    'text':'string',
    'float':'number',
    #from hemo
    'NUM':'number',
    'CHAR':'string'
}

formatmap = {
    'ISO8601':'' # NOTE: this is the default date format for frictionless so not necessary to specify
}

props = schemas.healjsonschema['properties']
    #mappings for array of dicts, arrays, and dicts


true_values = ["true","1","yes","required","y"]
false_values = ["false","0","no","not required","n"]

# cast numbers explicitly based on schema
# this is needed in case there is only one record in a string column that is a number (ie don't want to convert)
castnumbers = {
    field["name"]:int if field["type"]=="integer" else float
    for field in schemas.healcsvschema["fields"]
    if field.get("type","") in ["integer","number"]
}

fieldmap = {
    'constraints.required': lambda v: to_bool(v),
    'constraints.enum': lambda v: split_str_array(v),
    # 'constraints.maximum':int,
    # 'constraints.minimum':int, #TODO:need to add to schema
    # 'constraints.maxLength':int,
    'standardsMappings.type': lambda v: split_str_array(v),
    'standardsMappings.label': lambda v: split_str_array(v),
    'standardsMappings.source': lambda v: split_str_array(v),
    'standardsMappings.id': lambda v: split_str_array(v),
    'standardsMappings.url': lambda v: split_str_array(v),
    'relatedConcepts.type': lambda v: split_str_array(v),
    'relatedConcepts.label': lambda v: split_str_array(v),
    'relatedConcepts.source': lambda v: split_str_array(v),
    'relatedConcepts.id': lambda v: split_str_array(v),
    'relatedConcepts.url': lambda v: split_str_array(v),
    'encodings':lambda v: loads_dict(v),
    'format': lambda v: mapval(v,formatmap),
    'type':lambda v: mapval(v,typemap),
    #'univar_stats.cat_marginals':lambda v: split_and_map(v, prop['univar_stats']['cat_marginals']),
    'missingValues':lambda v: split_str_array(v),
    'trueValues': lambda v: split_str_array(v),
    'falseValues':lambda v: split_str_array(v),
    # TODO: add stats
}

zipmap = ["standardsMappings","relatedConcepts"]



""" 
functions to read and write files in a "smart" way
"""
import pyreadstat
import pandas as pd 
from pathlib import Path
import charset_normalizer
# read pyreadstat files

def read_pyreadstat(file_path,**kwargs):
    ''' 
    reads in a "metadata rich file"
    (dta, sav,b7bdat). Note, xport format not supported
    as it doesnt supply value labels.

    '''
    file_path = Path(file_path)
    ext = file_path.suffix 
    if ext=='.sav':
        read = pyreadstat.read_sav
    elif ext=='.sas7bdat':
        read = pyreadstat.read_sas7bdat
    elif ext=='.dta':
        read = pyreadstat.read_dta
    elif ext=='.por':
        read = pyreadstat.read_por

    return read(file_path,**kwargs)



def detect_file_encoding(file_path):
    """ 
    detects file encoding using charset_normalizer package
    """ 
    with open(file_path,'rb') as f:
        data = f.read()
        encoding_for_input = charset_normalizer.detect(data)

    is_confident = encoding_for_input["confidence"]==1
    if not is_confident:
        print("Be careful, the detected file encoding for:")
        print(f"{file_path}")
        print(r"has less than 100% confidence")
    #chardet_normalizer.detect returns confidence,encoding (as a string), and language (eg English)
    return encoding_for_input["encoding"]


def read_table(file_path,castdtype = "string"):
    """ 
    reads in a tabular file (ie spreadsheet) after detecting
    encoding and file extension without any type casting.

    currently supports csv and tsv

    defaults to not casting values (ie all columns are string dtypes)
    and not parsing strings into NA values (eg "" is kept as "")
    """ 
    ext = Path(file_path).suffix
    if ext==".csv":
        sep = ","
    elif ext==".tsv":
        sep = "\t"
        
    encoding = detect_file_encoding(file_path)
    file_encoding = pd.read_csv(
        file_path,sep=sep,encoding=encoding,dtype=castdtype,
        keep_default_na=False)

    return file_encoding





healcsvschema = {
    "description": "Variable level metadata individual fields integrated into "
    "the variable level\n"
    "metadata object within the HEAL platform metadata service.\n"
    "\n"
    "> Note, only `name` and `description` are required.\n"
    ">  Listed at the end of the description are suggested "
    '"priority" levels in brackets (e.g., [<priority>]):\n'
    "  1. [Required]: Needs to be filled out to be valid.\n"
    "  2. [Highly recommended]: Greatly help using the data "
    "dictionary but not required. \n"
    "  3. [Optional, if applicable]: May only be applicable to "
    "certain fields.\n"
    "  4. [Autopopulated, if not filled]: These fields are "
    "intended to be autopopulated from other fields but can be "
    "filled out if desired.\n"
    "  5. [Experimental]: These fields are not currently used but "
    "are in development.\n",
    "title": "HEAL Variable Level Metadata Fields",
    "fields": [
        {
            "name": "module",
            "description": "The section, form, survey instrument, set of "
            "measures  or other broad category used \n"
            "to group variables.\n",
            "title": "Module",
            "examples": [
                "Demographics",
                "PROMIS",
                "Substance use",
                "Medical History",
                "Sleep questions",
                "Physical activity",
            ],
            "type": "string",
        },
        {
            "name": "name",
            "description": "The name of a variable (i.e., field) as it "
            "appears in the data. \n"
            "\n"
            "[Required]\n",
            "title": "Variable Name",
            "type": "string",
            "constraints": {"required": True},
        },
        {
            "name": "title",
            "description": "The human-readable title or label of the "
            "variable. \n"
            "\n"
            "[Highly recommended]\n",
            "title": "Variable Label (ie Title)",
            "examples": ["My Variable (for name of my_variable)"],
            "type": "string",
        },
        {
            "name": "description",
            "description": "An extended description of the variable. This "
            "could be the definition of a variable or the \n"
            "question text (e.g., if a survey). \n"
            "\n"
            "[Required]\n",
            "title": "Variable Description",
            "examples": ["Definition", "Question text (if a survey)"],
            "type": "string",
            "constraints": {"required": True},
        },
        {
            "name": "type",
            "description": "A classification or category of a particular "
            "data element or property expected or allowed "
            "in the dataset.\n"
            "\n"
            "-  `number` (A numeric value with optional "
            "decimal places. (e.g., 3.14))\n"
            "-  `integer` (A whole number without decimal "
            "places. (e.g., 42))\n"
            "-  `string` (A sequence of characters. (e.g., "
            '\\"test\\"))\n'
            "-  `any` (Any type of data is allowed. (e.g., "
            "true))\n"
            "-  `boolean` (A binary value representing true "
            "or false. (e.g., true))\n"
            "-  `date` (A specific calendar date. (e.g., "
            '\\"2023-05-25\\"))\n'
            "-  `datetime` (A specific date and time, "
            "including timezone information. (e.g., "
            '\\"2023-05-25T10:30:00Z\\"))\n'
            "-  `time` (A specific time of day. (e.g., "
            '\\"10:30:00\\"))\n'
            "-  `year` (A specific year. (e.g., 2023)\n"
            "-  `yearmonth` (A specific year and month. "
            '(e.g., \\"2023-05\\"))\n'
            "-  `duration` (A length of time. (e.g., "
            '\\"PT1H\\")\n'
            "-  `geopoint` (A pair of latitude and "
            "longitude coordinates. (e.g., [51.5074, "
            "-0.1278]))\n",
            "title": "Variable Type",
            "type": "string",
            "constraints": {
                "enum": [
                    "integer",
                    "geopoint",
                    "string",
                    "yearmonth",
                    "number",
                    "date",
                    "boolean",
                    "any",
                    "year",
                    "duration",
                    "time",
                    "datetime",
                ]
            },
        },
        {
            "name": "format",
            "description": "A format taken from one of the [frictionless "
            "specification](https://specs.frictionlessdata.io/) "
            "schemas.\n"
            "For example, for tabular data, there is the "
            "[Table Schema "
            "specification](https://specs.frictionlessdata.io/table-schema)\n"
            "\n"
            "Each format is dependent on the `type` "
            "specified. For example:\n"
            'If `type` is "string", then see the String '
            "formats. \n"
            "If `type` is one of the date-like formats, "
            "then see Date formats.\n",
            "title": "Frictionless Formats",
            "type": "any",
        },
        {
            "name": "constraints.maxLength",
            "description": "Indicates the maximum length of an iterable "
            "(e.g., array, string, or\n"
            "object). For example, if 'Hello World' is the "
            "longest value of a\n"
            "categorical variable, this would be a "
            "maxLength of 11.\n"
            "\n"
            "[Optional,if applicable]\n",
            "title": "Maximum Length",
            "type": "integer",
        },
        {
            "name": "constraints.enum",
            "description": "Constrains possible values to a set of "
            "values.\n"
            "\n"
            "[Optional,if applicable]\n",
            "title": "Variable Possible Values",
            "type": "string",
            "constraints": {"pattern": "^(?:[^|]+\\||[^|]*)(?:[^|]*\\|)*[^|]*$"},
        },
        {
            "name": "constraints.pattern",
            "description": "A regular expression pattern the data MUST "
            "conform to.\n"
            "\n"
            "[Optional,if applicable]\n",
            "title": "Regular Expression Pattern",
            "type": "string",
        },
        {
            "name": "constraints.maximum",
            "description": "Specifies the maximum value of a field (e.g., "
            "maximum -- or most\n"
            "recent -- date, maximum integer etc). Note, "
            "this is different then\n"
            "maxLength property.\n"
            "\n"
            "[Optional,if applicable]\n",
            "title": "Maximum Value",
            "type": "integer",
        },
        {
            "name": "constraints.minimum",
            "description": "Specifies the minimum value of a field.\n"
            "\n"
            "[Optional,if applicable]\n",
            "title": "Minimum Value",
            "type": "integer",
        },
        {
            "name": "encodings",
            "description": "Variable value encodings provide a way to "
            "further annotate any value within a any "
            "variable type,\n"
            "making values easier to understand. \n"
            "\n"
            "\n"
            "Many analytic software programs (e.g., "
            "SPSS,Stata, and SAS) use numerical encodings "
            "and some algorithms\n"
            "only support numerical values. Encodings (and "
            "mappings) allow categorical values to be "
            "stored as\n"
            "numerical values.\n"
            "\n"
            "Additionally, as another use case, this field "
            "provides a way to\n"
            'store categoricals that are stored as  "short" '
            "labels (such as\n"
            "abbreviations).\n"
            "\n"
            "[Optional,if applicable]\n",
            "title": "Variable Value Encodings (i.e., mappings; value " "labels)",
            "examples": ["0=No|1=Yes", "HW=Hello world|GBW=Good bye world|HM=Hi,Mike"],
            "type": "string",
            "constraints": {"pattern": "^(?:.*?=.*?(?:\\||$))+$"},
        },
        {
            "name": "ordered",
            "description": "Indicates whether a categorical variable is "
            "ordered. This variable  is\n"
            "relevant for variables that have an ordered "
            "relationship but not\n"
            "necessarily  a numerical relationship (e.g., "
            "Strongly disagree < Disagree\n"
            "< Neutral < Agree).\n"
            "\n"
            "[Optional,if applicable]\n",
            "title": "An ordered variable",
            "type": "boolean",
        },
        {
            "name": "missingValues",
            "description": "A list of missing values specific to a "
            "variable.\n"
            "\n"
            "[Optional, if applicable]\n",
            "title": "Missing Values",
            "type": "string",
            "constraints": {"pattern": "^(?:[^|]+\\||[^|]*)(?:[^|]*\\|)*[^|]*$"},
        },
        {
            "name": "trueValues",
            "description": "For boolean (true) variable (as defined in "
            "type field), this field allows\n"
            "a physical string representation to be cast as "
            "true (increasing\n"
            "readability of the field). It can include one "
            "or more values.\n"
            "\n"
            "[Optional, if applicable]\n",
            "examples": [
                "Required|REQUIRED",
                "required|Yes|Y|Checked",
                "Checked",
                "Required",
            ],
            "type": "string",
            "constraints": {"pattern": "^(?:[^|]+\\||[^|]*)(?:[^|]*\\|)*[^|]*$"},
        },
        {
            "name": "falseValues",
            "description": "For boolean (false) variable (as defined in "
            "type field), this field allows\n"
            "a physical string representation to be cast as "
            "false (increasing\n"
            "readability of the field) that is not a "
            "standard false value. It can include one or "
            "more values.\n",
            "title": "Boolean False Value Labels",
            "type": "string",
            "constraints": {"pattern": "^(?:[^|]+\\||[^|]*)(?:[^|]*\\|)*[^|]*$"},
        },
        {
            "name": "repo_link",
            "description": "A link to the variable as it exists on the "
            "home repository, if applicable\n",
            "title": "Variable Repository Link",
            "type": "string",
        },
        {
            "name": "standardsMappings.type",
            "description": "The **type** of mapping linked to a published "
            "set of standard variables such as the NIH "
            "Common Data Elements program.\n"
            "[Autopopulated, if not filled]\n",
            "title": "Standards Mapping - Title",
            "examples": ["cde", "ontology", "reference_list"],
            "type": "string",
        },
        {
            "name": "standardsMappings.label",
            "description": "A free text **label** of a mapping indicating "
            "a mapping(s) to a published set of standard "
            "variables such as the NIH Common Data Elements "
            "program.\n"
            "\n"
            "[Autopopulated, if not filled]\n",
            "title": "Standards Mapping - Label",
            "examples": ["substance use", "chemical compound", "promis"],
            "type": "string",
        },
        {
            "name": "standardsMappings.url",
            "description": "The url that links out to the published, "
            "standardized mapping.\n"
            "\n"
            "[Autopopulated, if not filled]\n",
            "title": "Standards Mapping - Url",
            "examples": ["https://cde.nlm.nih.gov/deView?tinyId=XyuSGdTTI"],
            "type": "string",
        },
        {
            "name": "standardsMappings.source",
            "description": "The source of the standardized variable.\n",
            "title": "Standard Mapping - Source",
            "examples": ["TBD (will have controlled vocabulary)"],
            "type": "string",
        },
        {
            "name": "standardsMappings.id",
            "description": "The id locating the individual mapping within "
            "the given source.\n",
            "title": "Standard Mapping - Id",
            "type": "string",
        },
        {
            "name": "relatedConcepts.type",
            "description": "The **type** of mapping to a published set of "
            "concepts related to the given field such as \n"
            "ontological information (eg., NCI thesaurus, "
            "bioportal etc)\n"
            "\n"
            "[Autopopulated, if not filled]\n",
            "title": "Related concepts - Type",
            "type": "string",
        },
        {
            "name": "relatedConcepts.label",
            "description": "A free text **label** of mapping to a "
            "published set of concepts related to the given "
            "field such as \n"
            "ontological information (eg., NCI thesaurus, "
            "bioportal etc)\n"
            "\n"
            "[Autopopulated, if not filled]\n",
            "title": "Related Concepts - Label",
            "type": "string",
        },
        {
            "name": "relatedConcepts.url",
            "description": "The url that links out to the published, "
            "standardized concept.\n"
            "\n"
            "[Autopopulated, if not filled]\n",
            "title": "Related Concepts - Url",
            "examples": ["https://cde.nlm.nih.gov/deView?tinyId=XyuSGdTTI"],
            "type": "string",
        },
        {
            "name": "relatedConcepts.source",
            "description": "The source of the related concept.\n"
            "\n"
            "[Autopopulated, if not filled]\n",
            "title": "Related Concepts - Source",
            "examples": ["TBD (will have controlled vocabulary)"],
            "type": "string",
        },
        {
            "name": "relatedConcepts.id",
            "description": "The id locating the individual mapping within "
            "the given source.\n"
            "\n"
            "[Autopopulated, if not filled]\n",
            "title": "Related Concepts - Id",
            "type": "string",
        },
        {"name": "univarStats.median", "type": "number"},
        {"name": "univarStats.mean", "type": "number"},
        {"name": "univarStats.std", "type": "number"},
        {"name": "univarStats.min", "type": "number"},
        {"name": "univarStats.max", "type": "number"},
        {"name": "univarStats.mode", "type": "number"},
        {"name": "univarStats.count", "type": "integer"},
        {"name": "univarStats.twentyFifthPercentile", "type": "number"},
        {"name": "univarStats.seventyFifthPercentile", "type": "number"},
        {"name": "univarStats.categoricalMarginals.name", "type": "string"},
        {"name": "univarStats.categoricalMarginals.count", "type": "integer"},
    ],
    "missingValues": [""],
}



import requests

jsonschema_url = (
    "https://raw.githubusercontent.com/norc-heal/"
    "heal-metadata-schemas/mbkranz/variable-lvl-dev/"
    "variable-level-metadata-schema/schemas/jsonschema/data-dictionary.json"
)
csvschema_url = (
    "https://raw.githubusercontent.com/norc-heal/heal-metadata-schemas/"
    "mbkranz/variable-lvl-dev/"
    "variable-level-metadata-schema/schemas/frictionless/csvtemplate/fields.json"
)

# TODO: install schemas from pip or make it a submodule (so no GET call necessary for fields.json)
healjsonschema = requests.get(jsonschema_url).json()
healcsvschema = requests.get(csvschema_url).json()



healjsonschema = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "vlmd",
    "title": "Variable Level Metadata (Data Dictionaries)",
    "description": "This schema defines the variable level metadata for one data "
    "dictionary for a given study.Note a given study can have "
    "multiple data dictionaries",
    "type": "object",
    "required": ["title", "data_dictionary"],
    "properties": {
        "title": {"type": "string"},
        "description": {"type": "string"},
        "data_dictionary": {
            "type": "array",
            "items": {
                "$schema": "http://json-schema.org/draft-04/schema#",
                "$id": "vlmd-fields",
                "title": "HEAL Variable " "Level Metadata " "Fields",
                "description": "Variable "
                "level "
                "metadata "
                "individual "
                "fields "
                "integrated "
                "into the "
                "variable "
                "level\n"
                "metadata "
                "object "
                "within the "
                "HEAL "
                "platform "
                "metadata "
                "service.\n"
                "\n"
                "> Note, only "
                "`name` and "
                "`description` "
                "are "
                "required.\n"
                ">  Listed at "
                "the end of "
                "the "
                "description "
                "are "
                "suggested "
                '"priority" '
                "levels in "
                "brackets "
                "(e.g., "
                "[<priority>]):\n"
                "  1. "
                "[Required]: "
                "Needs to be "
                "filled out "
                "to be "
                "valid.\n"
                "  2. [Highly "
                "recommended]: "
                "Greatly help "
                "using the "
                "data "
                "dictionary "
                "but not "
                "required. \n"
                "  3. "
                "[Optional, "
                "if "
                "applicable]: "
                "May only be "
                "applicable "
                "to certain "
                "fields.\n"
                "  4. "
                "[Autopopulated, "
                "if not "
                "filled]: "
                "These fields "
                "are intended "
                "to be "
                "autopopulated "
                "from other "
                "fields but "
                "can be "
                "filled out "
                "if desired.\n"
                "  5. "
                "[Experimental]: "
                "These fields "
                "are not "
                "currently "
                "used but are "
                "in "
                "development.\n",
                "type": "object",
                "additionalProperties": True,
                "required": ["name", "description"],
                "properties": {
                    "module": {
                        "type": "string",
                        "title": "Module",
                        "description": "The "
                        "section, "
                        "form, "
                        "survey "
                        "instrument, "
                        "set "
                        "of "
                        "measures  "
                        "or "
                        "other "
                        "broad "
                        "category "
                        "used \n"
                        "to "
                        "group "
                        "variables.\n",
                        "examples": [
                            "Demographics",
                            "PROMIS",
                            "Substance " "use",
                            "Medical " "History",
                            "Sleep " "questions",
                            "Physical " "activity",
                        ],
                    },
                    "name": {
                        "type": "string",
                        "title": "Variable " "Name",
                        "description": "The "
                        "name "
                        "of "
                        "a "
                        "variable "
                        "(i.e., "
                        "field) "
                        "as "
                        "it "
                        "appears "
                        "in "
                        "the "
                        "data. \n"
                        "\n"
                        "[Required]\n",
                    },
                    "title": {
                        "type": "string",
                        "title": "Variable " "Label " "(ie " "Title)",
                        "description": "The "
                        "human-readable "
                        "title "
                        "or "
                        "label "
                        "of "
                        "the "
                        "variable. \n"
                        "\n"
                        "[Highly "
                        "recommended]\n",
                        "examples": [
                            "My " "Variable " "(for " "name " "of " "my_variable)"
                        ],
                    },
                    "description": {
                        "type": "string",
                        "title": "Variable " "Description",
                        "description": "An "
                        "extended "
                        "description "
                        "of "
                        "the "
                        "variable. "
                        "This "
                        "could "
                        "be "
                        "the "
                        "definition "
                        "of "
                        "a "
                        "variable "
                        "or "
                        "the \n"
                        "question "
                        "text "
                        "(e.g., "
                        "if "
                        "a "
                        "survey). \n"
                        "\n"
                        "[Required]\n",
                        "examples": [
                            "Definition",
                            "Question " "text " "(if " "a " "survey)",
                        ],
                    },
                    "type": {
                        "title": "Variable " "Type",
                        "description": "A "
                        "classification "
                        "or "
                        "category "
                        "of "
                        "a "
                        "particular "
                        "data "
                        "element "
                        "or "
                        "property "
                        "expected "
                        "or "
                        "allowed "
                        "in "
                        "the "
                        "dataset.\n"
                        "\n"
                        "-  "
                        "`number` "
                        "(A "
                        "numeric "
                        "value "
                        "with "
                        "optional "
                        "decimal "
                        "places. "
                        "(e.g., "
                        "3.14))\n"
                        "-  "
                        "`integer` "
                        "(A "
                        "whole "
                        "number "
                        "without "
                        "decimal "
                        "places. "
                        "(e.g., "
                        "42))\n"
                        "-  "
                        "`string` "
                        "(A "
                        "sequence "
                        "of "
                        "characters. "
                        "(e.g., "
                        '\\"test\\"))\n'
                        "-  "
                        "`any` "
                        "(Any "
                        "type "
                        "of "
                        "data "
                        "is "
                        "allowed. "
                        "(e.g., "
                        "true))\n"
                        "-  "
                        "`boolean` "
                        "(A "
                        "binary "
                        "value "
                        "representing "
                        "true "
                        "or "
                        "false. "
                        "(e.g., "
                        "true))\n"
                        "-  "
                        "`date` "
                        "(A "
                        "specific "
                        "calendar "
                        "date. "
                        "(e.g., "
                        '\\"2023-05-25\\"))\n'
                        "-  "
                        "`datetime` "
                        "(A "
                        "specific "
                        "date "
                        "and "
                        "time, "
                        "including "
                        "timezone "
                        "information. "
                        "(e.g., "
                        '\\"2023-05-25T10:30:00Z\\"))\n'
                        "-  "
                        "`time` "
                        "(A "
                        "specific "
                        "time "
                        "of "
                        "day. "
                        "(e.g., "
                        '\\"10:30:00\\"))\n'
                        "-  "
                        "`year` "
                        "(A "
                        "specific "
                        "year. "
                        "(e.g., "
                        "2023)\n"
                        "-  "
                        "`yearmonth` "
                        "(A "
                        "specific "
                        "year "
                        "and "
                        "month. "
                        "(e.g., "
                        '\\"2023-05\\"))\n'
                        "-  "
                        "`duration` "
                        "(A "
                        "length "
                        "of "
                        "time. "
                        "(e.g., "
                        '\\"PT1H\\")\n'
                        "-  "
                        "`geopoint` "
                        "(A "
                        "pair "
                        "of "
                        "latitude "
                        "and "
                        "longitude "
                        "coordinates. "
                        "(e.g., "
                        "[51.5074, "
                        "-0.1278]))\n",
                        "type": "string",
                        "enum": [
                            "number",
                            "integer",
                            "string",
                            "any",
                            "boolean",
                            "date",
                            "datetime",
                            "time",
                            "year",
                            "yearmonth",
                            "duration",
                            "geopoint",
                        ],
                    },
                    "format": {
                        "title": "Frictionless " "Formats",
                        "description": "A "
                        "format "
                        "taken "
                        "from "
                        "one "
                        "of "
                        "the "
                        "[frictionless "
                        "specification](https://specs.frictionlessdata.io/) "
                        "schemas.\n"
                        "For "
                        "example, "
                        "for "
                        "tabular "
                        "data, "
                        "there "
                        "is "
                        "the "
                        "[Table "
                        "Schema "
                        "specification](https://specs.frictionlessdata.io/table-schema)\n"
                        "\n"
                        "Each "
                        "format "
                        "is "
                        "dependent "
                        "on "
                        "the "
                        "`type` "
                        "specified. "
                        "For "
                        "example:\n"
                        "If "
                        "`type` "
                        "is "
                        '"string", '
                        "then "
                        "see "
                        "the "
                        "String "
                        "formats. \n"
                        "If "
                        "`type` "
                        "is "
                        "one "
                        "of "
                        "the "
                        "date-like "
                        "formats, "
                        "then "
                        "see "
                        "Date "
                        "formats.\n",
                        "name": "frictionless_formats",
                        "anyOf": [
                            {
                                "title": "String " "Format",
                                "enum": ["uri", "email", "binary", "uuid"],
                            },
                            {
                                "title": "Date " "Format",
                                "description": "A "
                                "format "
                                "for "
                                "a "
                                "date "
                                "variable "
                                "(`date`,`time`,`datetime`).  \n"
                                "    "
                                "\\n\\t* "
                                "**default**: "
                                "An "
                                "ISO8601 "
                                "format "
                                "string.\n"
                                "    "
                                "\\n\\t* "
                                "**any**: "
                                "Any "
                                "parsable "
                                "representation "
                                "of "
                                "a "
                                "date/time/datetime. "
                                "The "
                                "implementing "
                                "library "
                                "can "
                                "attempt "
                                "to "
                                "parse "
                                "the "
                                "datetime "
                                "via "
                                "a "
                                "range "
                                "of "
                                "strategies.\n"
                                "    "
                                "\\n\\t* "
                                "**{PATTERN}**: "
                                "The "
                                "value "
                                "can "
                                "be "
                                "parsed "
                                "according "
                                "to "
                                "`{PATTERN}`, "
                                "which "
                                "`MUST` "
                                "follow "
                                "the "
                                "date "
                                "formatting "
                                "syntax "
                                "of "
                                "C "
                                "/ "
                                "Python "
                                "[strftime](http://strftime.org/).\n"
                                "\n"
                                "\\nExamples:\n"
                                "\n"
                                "  "
                                "`%Y-%m-%d` "
                                "(for "
                                "date, "
                                "e.g., "
                                "2023-05-25)\n"
                                "  "
                                "`%Y%-%d` "
                                "(for "
                                "date, "
                                "e.g., "
                                "20230525) "
                                "for "
                                "date "
                                "without "
                                'dashes"\n'
                                "  "
                                "`%Y-%m-%dT%H:%M:%S` "
                                "(for "
                                "datetime, "
                                "e.g., "
                                "2023-05-25T10:30:45)\n"
                                "  "
                                "`%Y-%m-%dT%H:%M:%SZ` "
                                "(for "
                                "datetime "
                                "with "
                                "UTC "
                                "timezone, "
                                "e.g., "
                                "2023-05-25T10:30:45Z)\n"
                                "  "
                                "`%Y-%m-%dT%H:%M:%S%z` "
                                "(for "
                                "datetime "
                                "with "
                                "timezone "
                                "offset, "
                                "e.g., "
                                "2023-05-25T10:30:45+0300)\n"
                                "  "
                                "`%Y-%m-%dT%H:%M` "
                                "(for "
                                "datetime "
                                "without "
                                "seconds, "
                                "e.g., "
                                "2023-05-25T10:30)\n"
                                "  "
                                "`%Y-%m-%dT%H` "
                                "(for "
                                "datetime "
                                "without "
                                "minutes "
                                "and "
                                "seconds, "
                                "e.g., "
                                "2023-05-25T10)\n"
                                "  "
                                "`%H:%M:%S` "
                                "(for "
                                "time, "
                                "e.g., "
                                "10:30:45)\n"
                                "  "
                                "`%H:%M:%SZ` "
                                "(for "
                                "time "
                                "with "
                                "UTC "
                                "timezone, "
                                "e.g., "
                                "10:30:45Z)\n"
                                "  "
                                "`%H:%M:%S%z` "
                                "(for "
                                "time "
                                "with "
                                "timezone "
                                "offset, "
                                "e.g., "
                                "10:30:45+0300)\n",
                            },
                            {
                                "title": "Geopoint " "Format",
                                "description": "The "
                                "two "
                                "types "
                                "of "
                                "formats "
                                "for "
                                "`geopoint` "
                                "(describing "
                                "a "
                                "geographic "
                                "point).",
                                "oneOf": [
                                    {
                                        "type": "array",
                                        "description": "A "
                                        "JSON "
                                        "array "
                                        "or "
                                        "a "
                                        "string "
                                        "parsable "
                                        "as "
                                        "a "
                                        "JSON "
                                        "array "
                                        "where "
                                        "each "
                                        "item "
                                        "is "
                                        "a "
                                        "number "
                                        "with "
                                        "the "
                                        "first \n"
                                        "as "
                                        "the "
                                        "latitude "
                                        "and "
                                        "the "
                                        "second "
                                        "as "
                                        "longitude. \n",
                                    },
                                    {
                                        "type": "object",
                                        "description": "Contains "
                                        "latitude "
                                        "and "
                                        "longitude "
                                        "with "
                                        "two "
                                        "keys "
                                        '("lat" '
                                        "and "
                                        '"long") '
                                        "with "
                                        "number "
                                        "items "
                                        "mapped "
                                        "to "
                                        "each "
                                        "key.\n",
                                    },
                                ],
                            },
                            {
                                "title": "geojson",
                                "description": "The "
                                "JSON "
                                "object "
                                "according "
                                "to "
                                "the "
                                "geojson "
                                "spec.",
                                "enum": ["topojson", "default"],
                            },
                        ],
                    },
                    "constraints": {
                        "type": "object",
                        "properties": {
                            "maxLength": {
                                "type": "integer",
                                "title": "Maximum " "Length",
                                "description": "Indicates "
                                "the "
                                "maximum "
                                "length "
                                "of "
                                "an "
                                "iterable "
                                "(e.g., "
                                "array, "
                                "string, "
                                "or\n"
                                "object). "
                                "For "
                                "example, "
                                "if "
                                "'Hello "
                                "World' "
                                "is "
                                "the "
                                "longest "
                                "value "
                                "of "
                                "a\n"
                                "categorical "
                                "variable, "
                                "this "
                                "would "
                                "be "
                                "a "
                                "maxLength "
                                "of "
                                "11.\n"
                                "\n"
                                "[Optional,if "
                                "applicable]\n",
                            },
                            "enum": {
                                "title": "Variable " "Possible " "Values",
                                "description": "Constrains "
                                "possible "
                                "values "
                                "to "
                                "a "
                                "set "
                                "of "
                                "values.\n"
                                "\n"
                                "[Optional,if "
                                "applicable]\n",
                                "type": "array",
                            },
                            "pattern": {
                                "type": "string",
                                "title": "Regular " "Expression " "Pattern",
                                "description": "A "
                                "regular "
                                "expression "
                                "pattern "
                                "the "
                                "data "
                                "MUST "
                                "conform "
                                "to.\n"
                                "\n"
                                "[Optional,if "
                                "applicable]\n",
                            },
                            "maximum": {
                                "type": "integer",
                                "title": "Maximum " "Value",
                                "description": "Specifies "
                                "the "
                                "maximum "
                                "value "
                                "of "
                                "a "
                                "field "
                                "(e.g., "
                                "maximum "
                                "-- "
                                "or "
                                "most\n"
                                "recent "
                                "-- "
                                "date, "
                                "maximum "
                                "integer "
                                "etc). "
                                "Note, "
                                "this "
                                "is "
                                "different "
                                "then\n"
                                "maxLength "
                                "property.\n"
                                "\n"
                                "[Optional,if "
                                "applicable]\n",
                            },
                            "minimum": {
                                "type": "integer",
                                "title": "Minimum " "Value",
                                "description": "Specifies "
                                "the "
                                "minimum "
                                "value "
                                "of "
                                "a "
                                "field.\n"
                                "\n"
                                "[Optional,if "
                                "applicable]\n",
                            },
                        },
                    },
                    "encodings": {
                        "title": "Variable "
                        "Value "
                        "Encodings "
                        "(i.e., "
                        "mappings; "
                        "value "
                        "labels)",
                        "description": "Variable "
                        "value "
                        "encodings "
                        "provide "
                        "a "
                        "way "
                        "to "
                        "further "
                        "annotate "
                        "any "
                        "value "
                        "within "
                        "a "
                        "any "
                        "variable "
                        "type,\n"
                        "making "
                        "values "
                        "easier "
                        "to "
                        "understand. \n"
                        "\n"
                        "\n"
                        "Many "
                        "analytic "
                        "software "
                        "programs "
                        "(e.g., "
                        "SPSS,Stata, "
                        "and "
                        "SAS) "
                        "use "
                        "numerical "
                        "encodings "
                        "and "
                        "some "
                        "algorithms\n"
                        "only "
                        "support "
                        "numerical "
                        "values. "
                        "Encodings "
                        "(and "
                        "mappings) "
                        "allow "
                        "categorical "
                        "values "
                        "to "
                        "be "
                        "stored "
                        "as\n"
                        "numerical "
                        "values.\n"
                        "\n"
                        "Additionally, "
                        "as "
                        "another "
                        "use "
                        "case, "
                        "this "
                        "field "
                        "provides "
                        "a "
                        "way "
                        "to\n"
                        "store "
                        "categoricals "
                        "that "
                        "are "
                        "stored "
                        "as  "
                        '"short" '
                        "labels "
                        "(such "
                        "as\n"
                        "abbreviations).\n"
                        "\n"
                        "[Optional,if "
                        "applicable]\n",
                        "type": "object",
                        "examples": [
                            {"0": "No", "1": "Yes"},
                            {
                                "HW": "Hello " "world",
                                "GBW": "Good " "bye " "world",
                                "HM": "Hi, " "Mike",
                            },
                        ],
                    },
                    "ordered": {
                        "title": "An " "ordered " "variable",
                        "description": "Indicates "
                        "whether "
                        "a "
                        "categorical "
                        "variable "
                        "is "
                        "ordered. "
                        "This "
                        "variable  "
                        "is\n"
                        "relevant "
                        "for "
                        "variables "
                        "that "
                        "have "
                        "an "
                        "ordered "
                        "relationship "
                        "but "
                        "not\n"
                        "necessarily  "
                        "a "
                        "numerical "
                        "relationship "
                        "(e.g., "
                        "Strongly "
                        "disagree "
                        "< "
                        "Disagree\n"
                        "< "
                        "Neutral "
                        "< "
                        "Agree).\n"
                        "\n"
                        "[Optional,if "
                        "applicable]\n",
                        "type": "boolean",
                    },
                    "missingValues": {
                        "title": "Missing " "Values",
                        "description": "A "
                        "list "
                        "of "
                        "missing "
                        "values "
                        "specific "
                        "to "
                        "a "
                        "variable.\n"
                        "\n"
                        "[Highly "
                        "recommended]\n",
                        "type": "array",
                    },
                    "trueValues": {
                        "title": "Boolean " "True " "Value " "Labels",
                        "description": "For "
                        "boolean "
                        "(true) "
                        "variable "
                        "(as "
                        "defined "
                        "in "
                        "type "
                        "field), "
                        "this "
                        "field "
                        "allows\n"
                        "a "
                        "physical "
                        "string "
                        "representation "
                        "to "
                        "be "
                        "cast "
                        "as "
                        "true "
                        "(increasing\n"
                        "readability "
                        "of "
                        "the "
                        "field). "
                        "It "
                        "can "
                        "include "
                        "one "
                        "or "
                        "more "
                        "values.\n"
                        "\n"
                        "[Optional, "
                        "if "
                        "applicable]\n",
                        "type": "array",
                        "items": {"type": "string"},
                        "examples": [
                            "Required",
                            "REQUIRED",
                            "required",
                            "Yes",
                            'Checked"',
                        ],
                    },
                    "falseValues": {
                        "title": "Boolean " "False " "Value " "Labels",
                        "description": "For "
                        "boolean "
                        "(false) "
                        "variable "
                        "(as "
                        "defined "
                        "in "
                        "type "
                        "field), "
                        "this "
                        "field "
                        "allows\n"
                        "a "
                        "physical "
                        "string "
                        "representation "
                        "to "
                        "be "
                        "cast "
                        "as "
                        "false "
                        "(increasing\n"
                        "readability "
                        "of "
                        "the "
                        "field) "
                        "that "
                        "is "
                        "not "
                        "a "
                        "standard "
                        "false "
                        "value. "
                        "It "
                        "can "
                        "include "
                        "one "
                        "or "
                        "more "
                        "values.\n",
                        "type": "array",
                    },
                    "repo_link": {
                        "type": "string",
                        "title": "Variable " "Repository " "Link",
                        "description": "A "
                        "link "
                        "to "
                        "the "
                        "variable "
                        "as "
                        "it "
                        "exists "
                        "on "
                        "the "
                        "home "
                        "repository, "
                        "if "
                        "applicable\n",
                    },
                    "standardsMappings": {
                        "title": "Standards " "Mappings",
                        "description": "A "
                        "published "
                        "set "
                        "of "
                        "standard "
                        "variables "
                        "such "
                        "as "
                        "the "
                        "NIH "
                        "Common "
                        "Data "
                        "Elements "
                        "program.\n"
                        "[Autopopulated, "
                        "if "
                        "not "
                        "filled]",
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "type": {
                                    "title": "Standards " "Mapping " "- " "Title",
                                    "description": "The "
                                    "**type** "
                                    "of "
                                    "mapping "
                                    "linked "
                                    "to "
                                    "a "
                                    "published "
                                    "set "
                                    "of "
                                    "standard "
                                    "variables "
                                    "such "
                                    "as "
                                    "the "
                                    "NIH "
                                    "Common "
                                    "Data "
                                    "Elements "
                                    "program.\n"
                                    "[Autopopulated, "
                                    "if "
                                    "not "
                                    "filled]\n",
                                    "examples": ["cde", "ontology", "reference_list"],
                                    "type": "string",
                                },
                                "label": {
                                    "title": "Standards " "Mapping " "- " "Label",
                                    "description": "A "
                                    "free "
                                    "text "
                                    "**label** "
                                    "of "
                                    "a "
                                    "mapping "
                                    "indicating "
                                    "a "
                                    "mapping(s) "
                                    "to "
                                    "a "
                                    "published "
                                    "set "
                                    "of "
                                    "standard "
                                    "variables "
                                    "such "
                                    "as "
                                    "the "
                                    "NIH "
                                    "Common "
                                    "Data "
                                    "Elements "
                                    "program.\n"
                                    "\n"
                                    "[Autopopulated, "
                                    "if "
                                    "not "
                                    "filled]\n",
                                    "type": "string",
                                    "examples": [
                                        "substance " "use",
                                        "chemical " "compound",
                                        "promis",
                                    ],
                                },
                                "url": {
                                    "title": "Standards " "Mapping " "- " "Url",
                                    "description": "The "
                                    "url "
                                    "that "
                                    "links "
                                    "out "
                                    "to "
                                    "the "
                                    "published, "
                                    "standardized "
                                    "mapping.\n"
                                    "\n"
                                    "[Autopopulated, "
                                    "if "
                                    "not "
                                    "filled]\n",
                                    "type": "string",
                                    "format": "uri",
                                    "examples": [
                                        "https://cde.nlm.nih.gov/deView?tinyId=XyuSGdTTI"
                                    ],
                                },
                                "source": {
                                    "title": "Standard " "Mapping " "- " "Source",
                                    "description": "The "
                                    "source "
                                    "of "
                                    "the "
                                    "standardized "
                                    "variable.\n",
                                    "type": "string",
                                    "examples": [
                                        "TBD "
                                        "(will "
                                        "have "
                                        "controlled "
                                        "vocabulary)"
                                    ],
                                },
                                "id": {
                                    "title": "Standard " "Mapping " "- " "Id",
                                    "type": "string",
                                    "description": "The "
                                    "id "
                                    "locating "
                                    "the "
                                    "individual "
                                    "mapping "
                                    "within "
                                    "the "
                                    "given "
                                    "source.\n",
                                },
                            },
                        },
                    },
                    "relatedConcepts": {
                        "title": "Related " "Concepts",
                        "description": "Mappings "
                        "to "
                        "a "
                        "published "
                        "set "
                        "of "
                        "concepts "
                        "related "
                        "to "
                        "the "
                        "given "
                        "field "
                        "such "
                        "as "
                        "ontological "
                        "information "
                        "(eg., "
                        "NCI "
                        "thesaurus, "
                        "bioportal "
                        "etc)\n"
                        "[Autopopulated, "
                        "if "
                        "not "
                        "filled]",
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "type": {
                                    "title": "Related " "concepts " "- " "Type",
                                    "description": "The "
                                    "**type** "
                                    "of "
                                    "mapping "
                                    "to "
                                    "a "
                                    "published "
                                    "set "
                                    "of "
                                    "concepts "
                                    "related "
                                    "to "
                                    "the "
                                    "given "
                                    "field "
                                    "such "
                                    "as \n"
                                    "ontological "
                                    "information "
                                    "(eg., "
                                    "NCI "
                                    "thesaurus, "
                                    "bioportal "
                                    "etc)\n"
                                    "\n"
                                    "[Autopopulated, "
                                    "if "
                                    "not "
                                    "filled]\n",
                                    "type": "string",
                                },
                                "label": {
                                    "type": "string",
                                    "title": "Related " "Concepts " "- " "Label",
                                    "description": "A "
                                    "free "
                                    "text "
                                    "**label** "
                                    "of "
                                    "mapping "
                                    "to "
                                    "a "
                                    "published "
                                    "set "
                                    "of "
                                    "concepts "
                                    "related "
                                    "to "
                                    "the "
                                    "given "
                                    "field "
                                    "such "
                                    "as \n"
                                    "ontological "
                                    "information "
                                    "(eg., "
                                    "NCI "
                                    "thesaurus, "
                                    "bioportal "
                                    "etc)\n"
                                    "\n"
                                    "[Autopopulated, "
                                    "if "
                                    "not "
                                    "filled]\n",
                                },
                                "url": {
                                    "title": "Related " "Concepts " "- " "Url",
                                    "description": "The "
                                    "url "
                                    "that "
                                    "links "
                                    "out "
                                    "to "
                                    "the "
                                    "published, "
                                    "standardized "
                                    "concept.\n"
                                    "\n"
                                    "[Autopopulated, "
                                    "if "
                                    "not "
                                    "filled]\n",
                                    "type": "string",
                                    "format": "uri",
                                    "examples": [
                                        "https://cde.nlm.nih.gov/deView?tinyId=XyuSGdTTI"
                                    ],
                                },
                                "source": {
                                    "title": "Related " "Concepts " "- " "Source",
                                    "description": "The "
                                    "source "
                                    "of "
                                    "the "
                                    "related "
                                    "concept.\n"
                                    "\n"
                                    "[Autopopulated, "
                                    "if "
                                    "not "
                                    "filled]\n",
                                    "type": "string",
                                    "examples": [
                                        "TBD "
                                        "(will "
                                        "have "
                                        "controlled "
                                        "vocabulary)"
                                    ],
                                },
                                "id": {
                                    "title": "Related " "Concepts " "- " "Id",
                                    "type": "string",
                                    "description": "The "
                                    "id "
                                    "locating "
                                    "the "
                                    "individual "
                                    "mapping "
                                    "within "
                                    "the "
                                    "given "
                                    "source.\n"
                                    "\n"
                                    "[Autopopulated, "
                                    "if "
                                    "not "
                                    "filled]\n",
                                },
                            },
                        },
                    },
                    "univarStats": {
                        "type": "object",
                        "description": "Univariate "
                        "statistics "
                        "inferred "
                        "from "
                        "the "
                        "data "
                        "about "
                        "the "
                        "given "
                        "variable \n"
                        "\n"
                        "[Experimental]\n",
                        "properties": {
                            "median": {"type": "number"},
                            "mean": {"type": "number"},
                            "std": {"type": "number"},
                            "min": {"type": "number"},
                            "max": {"type": "number"},
                            "mode": {"type": "number"},
                            "count": {"type": "integer", "minimum": 0},
                            "twentyFifthPercentile": {"type": "number"},
                            "seventyFifthPercentile": {"type": "number"},
                            "categoricalMarginals": {
                                "type": "array",
                                "items": {
                                    "type": "object",
                                    "properties": {
                                        "name": {"type": "string"},
                                        "count": {"type": "integer"},
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
    },
}



from pathlib import Path
import json
# from frictionless import Resource,Package
from collections.abc import MutableMapping
from .mappings import join_prop
from healdata_utils.utils import flatten_except_if
from os import PathLike

def convert_templatejson(
    jsontemplate,
    data_dictionary_props:dict=None,
    fields_name:str='data_dictionary',
    sep_iter = '|',
    sep_dict = '=',
    **kwargs
    ):
    """
    Converts a JSON file or dictionary conforming to HEAL specifications
    into a HEAL-specified data dictionary in both csv format and json format.

    Converts in-memory data or a path to a data dictionary file.

    If data_dictionary_props is specified, any properties passed in will be
    overwritten.
    
    Parameters
    ----------
    jsontemplate : str or path-like or an object that can be inferred as data by frictionless's Resource class.
        Data or path to data with the data being a tabular HEAL-specified data dictionary.
        This input can be any data object or path-like string excepted by a frictionless Resource object.
    data_dictionary_props : dict
        The HEAL-specified data dictionary properties.
    mappings : dict, optional
        Mappings (which can be a dictionary of either lambda functions or other to-be-mapped objects).
        Default: specified fieldmap.

    Returns
    -------
    dict
        A dictionary with two keys:
            - 'templatejson': the HEAL-specified JSON object.
            - 'templatecsv': the HEAL-specified tabular template.


    TODO
    ---------

    Allow an array of fields to be passed in

    """
    if isinstance(jsontemplate,(str,PathLike)):
        jsontemplate_dict = json.loads(Path(jsontemplate).read_text())
    elif isinstance(jsontemplate, MutableMapping):
        jsontemplate_dict = jsontemplate
    else:
        raise Exception("jsontemplate needs to be either dictionary-like or a path to a json")

    if data_dictionary_props:
        for propname,prop in data_dictionary_props.items():

            # determine if you should write or overwrite the
            ## root level data dictionary props
            if not jsontemplate_dict.get(propname):
                write_prop = True
            elif prop and prop!=jsontemplate_dict.get(propname):
                write_prop = True
            else:
                write_prop = False

            if write_prop:
                jsontemplate_dict[propname] = prop

    fields_json = jsontemplate_dict.pop(fields_name)
    data_dictionary_props = jsontemplate_dict
    
    fields_csv = []
    for f in fields_json:
        field_flattened = flatten_except_if(f)
        field_csv = {
            propname:join_prop(propname,prop)
            for propname,prop in field_flattened.items()
        }
        fields_csv.append(field_csv)

    template_json = dict(**data_dictionary_props,data_dictionary=fields_json)
    template_csv = dict(**data_dictionary_props,data_dictionary=fields_csv)

    return {"templatejson":template_json,"templatecsv":template_csv}


import shutil
from pathlib import Path
from healdata_utils.cli import convert_to_vlmd
import json

def test_convert_to_vlmd_with_registered_formats(
    valid_input_params,valid_output_json,valid_output_csv,fields_propname):
    inputtypes = list(valid_input_params.keys())
    outputdir="tmp"

    for inputtype in inputtypes:
   
        # make an empty temporary output directory
        try:
            Path(outputdir).mkdir()
        except FileExistsError:
            shutil.rmtree(outputdir)
            Path(outputdir).mkdir()

        _valid_input_params = valid_input_params[inputtype]
        _valid_output_json = valid_output_json[inputtype]
        _valid_output_csv = valid_output_csv[inputtype]

        data_dictionaries = convert_to_vlmd(
            **_valid_input_params,
        outputdir=outputdir
        )

        ddjson = json.loads(list(Path("tmp").glob("*.json"))[0].read_text())
        #NOTE: csv are just fields so no ddcsv

        # check for incorrect fields       
        csv_fields = list(Path("tmp").glob("*.csv"))[0].read_text().split("\n")
        json_fields = ddjson.pop(fields_propname) #NOTE: testing individual fields

        valid_output_json_fields = _valid_output_json.pop(fields_propname)
        valid_output_csv_fields = _valid_output_csv

        invalid_json_fields = []
        invalid_csv_fields = []
        indices = range(len(json_fields))
        for i in indices:
            if json_fields[i]!=valid_output_json_fields[i]:
                invalid_json_fields.append(i)
            if csv_fields[i]!=valid_output_csv_fields[i]:
                invalid_csv_fields.append(i)
        
        json_field_names = [f["name"] for f in json_fields]
        csv_field_names = [f["name"] for f in json_fields]

        assert sorted(json_field_names)==sorted(csv_field_names),f"{inputtype} conversion: json fields must have the same field names as csv fields"
        assert len(invalid_json_fields)==0,f"{inputtype} conversion: The following **json** dd fields are not valid: {str(invalid_json_fields)}"
        assert len(invalid_csv_fields)==0,f"{inputtype} conversion: The following **csv** dd fields are not valid: {str(invalid_csv_fields)}"
        
        
         # check if root level properties other than the fields are valid
        for propname in ddjson:
            assert ddjson[propname] == _valid_output_json[propname],f"{inputtype} conversion to json dd property '{propname}' assertion failed"

    
        # clean up
        shutil.rmtree(outputdir)



''' 

command line interface for generating HEAL data dictionary/vlmd json files

''' 

import click 
from healdata_utils.transforms.csvtemplate.conversion import convert_templatecsv
from healdata_utils.transforms.jsontemplate.conversion import convert_templatejson
from healdata_utils.transforms.readstat.conversion import convert_readstat
from healdata_utils.transforms.redcapcsv.conversion import convert_redcapcsv
from healdata_utils.transforms.csvdata.conversion import convert_datacsv
from healdata_utils.transforms.frictionless.conversion import convert_frictionless_tableschema

from healdata_utils.validators.validate import validate_vlmd_json,validate_vlmd_csv
import json
from pathlib import Path
import petl as etl
import csv
from collections import deque

from healdata_utils.utils import find_docstring_desc

choice_fxn = {
    'data.csv':convert_datacsv,
    'template.csv':convert_templatecsv,
    'csv': convert_templatecsv, #maintained for backwards compatibility
    'sav': convert_readstat,
    'dta':convert_readstat,
    #'por':convert_readstat,
    'sas7bdat':convert_readstat,
    'template.json':convert_templatejson,
    'json':convert_templatejson, #maintain for bwds compat
    "redcap.csv":convert_redcapcsv,
    "frictionless.schema.json":convert_frictionless_tableschema

}

input_types = " - "+"\n - ".join(list(choice_fxn.keys()))

input_descriptions = {
    name:find_docstring_desc(fxn)
    for name,fxn in choice_fxn.items()
}

def convert_to_vlmd(
    filepath,
    data_dictionary_props={},
    inputtype=None,
    outputdir=None,
    sas7bcat_filepath=None,
    csvtemplate_output_quoting=None,

    ):
    """
    Writes a data dictionary (i.e. variable level metadata) to a HEAL metadata JSON file using a registered function.

    Parameters
    ----------
    filepath : str
        Path to input file. See documentation on individual input types for more details.
    outputdir : str
        output file path or directory to where output will go. 
        1. Outputdir is a directory, will give standard names,
        2. If path to the to-be-written file is specified, will use this as new outputted filename
    data_dictionary_props : dict, optional
        The other data-dictionary level properties. By default, will give the data_dictionary `title` property as the file name stem.
    inputtype : str, optional
        The input type. If none specified, will default to using the file extension.
        See the currently registered input types in the input_types list.
    sas7bcat_filepath: str,optional
        [FOR SAS7BDAT ONLY]: Path to a sas catalog file (sas7bcat). Needed for value formats if a sas (sas7bdat) input file
    csvtemplate_output_quoting: bool, optional
        If true, all nonnumeric values will be quoted. This helps reduce ambiguity for programs
        like excel that uses special characters for specific purposes (eg = for formulas)
    
    Returns
    -------
    dict
        Dictionary with:
         1. csvtemplated array of fields.
         2. jsontemplated data dictionary object as specified by an originally drafted design doc.
            That is, a dictionary with title:<title>,description:<description>,data_dictionary:<fields>
            where data dictionary is an array of fields as specified by the JSON schema.
         3. error objects for corresponding validators (ie frictionless for csv and jsonschema for json)
    NOTE
    ----
    In future versions, this will be more of a package bundled with corresponding schemas (whether csv or JSON),better organization 
    (e.g., see frictionless Package). 
    However, right now, it simply returns the csvtemplate and jsontemplate as specified
    in the heal specification repository.
    This is an intermediate solution to socialize a proof-of-concept.

    TODO
    --------
    make sub command for each file format rather than just one function? Added sas7bcat and can predict additional complexity
    
    """

    filepath = Path(filepath)
    
    #infer input type
    if not inputtype:
        inputtype = ''.join(filepath.suffixes)[1:].lower()

    ## add dd title
    if not data_dictionary_props.get('title'):
        data_dictionary_props['title'] = filepath.stem

    
    # get data dictionary package based on the input type
    if sas7bcat_filepath:
        data_dictionary_package = choice_fxn[inputtype](filepath,data_dictionary_props,sas7bcat_file_path=sas7bcat_filepath,) 
    else:
        data_dictionary_package = choice_fxn[inputtype](filepath,data_dictionary_props)


    #TODO: currently only validates fields (ie table) but no reason it cant validate entire data package
    package_csv = validate_vlmd_csv(data_dictionary_package['templatecsv']['data_dictionary'],to_sync_fields=True)
    package_json = validate_vlmd_json(data_dictionary_package['templatejson'])

    # TODO: in future just return the packages (eg reports nested within package and not out of)
    # for now, keep same (report_xxx and templatexxx)

    report_csv = package_csv["report"]
    report_json = package_json["report"]
    templatecsv = package_csv["data"]
    templatejson = package_json["data"]

    # write to file
    if outputdir!=None:
        outputdir = Path(outputdir)
        if outputdir.is_dir():
            jsontemplate_path = outputdir/"heal-jsontemplate-data-dictionary.json"
            csvtemplate_path = outputdir/"heal-csvtemplate-data-dictionary.csv"
        elif outputdir.parent.is_dir():
            jsontemplate_path = outputdir.with_suffix(".json")
            csvtemplate_path = outputdir.with_suffix(".csv")
            outputdir = outputdir.parent
        else:
            raise Exception("outputdir must be an existing directory where files can be saved")
        
        # print data dictionaries
        jsontemplate_path.write_text(json.dumps(templatejson,indent=4))

        # NOTE: quoting non-numeric to allow special characters for nested delimiters within string columns (ie "=")
        (
            etl.fromdicts(templatecsv)
            .tocsv(
                csvtemplate_path,
                quoting=csv.QUOTE_NONNUMERIC if csvtemplate_output_quoting else csv.QUOTE_MINIMAL)

        )

        # print errors

        if not report_json['valid']:
            print("JSON data dictionary not valid, see heal-json-errors.json for errors.")
            print(f"(view the outputted data dictionary at {jsontemplate_path})")
 
        if not report_csv['valid']:
            print("CSV data dictionary not valid, see heal-csv-errors.json")
            print(f"(view the outputted data dictionary at {csvtemplate_path})")
        
        # write error reports to file
        errordir = Path(outputdir).joinpath('errors')
        errordir.mkdir(exist_ok=True)
        errordir.joinpath('heal-json-errors.json').write_text(
            json.dumps(report_json,indent=4)
        )
        errordir.joinpath('heal-csv-errors.json').write_text(
            json.dumps(report_csv,indent=4)
        )
    
    return {
        "csvtemplate":templatecsv,
        "jsontemplate":templatejson,
        "errors":{
            "csvtemplate":report_csv,
            "jsontemplate":report_json}
        }

@click.command()
@click.option('--filepath',required=True,help='Path to the file you want to convert to a HEAL data dictionary')
@click.option('--title',default=None,help='The title of your data dictionary. If not specified, then the file name will be used')
@click.option('--description',default=None,help='Description of data dictionary')
@click.option('--inputtype',default=None,type=click.Choice(list(choice_fxn.keys())),help='The type of your input file.')
@click.option('--outputdir',default="",help='The folder where you want to output your HEAL data dictionary')
@click.option('--sas7bcat-filepath',default=None,help="[FOR SAS7BDAT ONLY]: Path to a sas catalog file (sas7bcat). Needed for value formats if a sas (sas7bdat) input file")
@click.option('--csvtemplate-output-quoting',default=None,help="If true, all nonnumeric values will be quoted."
    " This helps reduce ambiguity for programs like excel that uses special characters for specific purposes (eg = for formulas)")
def main(filepath,title,description,inputtype,outputdir,sas7bcat_filepath,csvtemplate_output_quoting):
    data_dictionary_props = {'title':title,'description':description}

    #save dds and error reports to files
    data_dictionaries = convert_to_vlmd(
        filepath=filepath,
        data_dictionary_props=data_dictionary_props,
        outputdir=outputdir,
        inputtype=inputtype,
        sas7bcat_filepath=sas7bcat_filepath,
        csvtemplate_output_quoting=csvtemplate_output_quoting
    )
     
if __name__=='__main__':
    main()